import typing
from dataclasses import field, dataclass
from typing import Protocol


TVar = typing.TypeVar('TVar')


@dataclass(unsafe_hash=True)
class Error:
    message: str = None


@dataclass(unsafe_hash=True)
class ApplicationContext:
    body: dict = None
    auth_user_id: str = None
    variables: dict = None
    error_capsules: list[Error] = field(default_factory=lambda: [])
    response: typing.Any = None

    def get_var(self, name: str, _type: typing.Type[TVar]) -> TVar:
        return self.variables[name]

    def set_var(self, name: str, value: typing.Any):
        self.variables[name] = value


class ErrorHandlingStrategy(Protocol):

    def handle_error(self, context: ApplicationContext, error: Error) -> None:
        ...

    @property
    def strategy(self) -> str:
        ...


class Command(Protocol):

    def run(self, context: ApplicationContext) -> None:
        ...


class ApiRequestHandler(Protocol):

    def run(self, event: dict) -> ApplicationContext:
        ...

    @property
    def route_key(self) -> str:
        ...


class EventHandler(Protocol):

    def run(self, event: str) -> ApplicationContext:
        ...

    @property
    def event_type(self) -> typing.Type:
        ...


class SequenceBuilder(Protocol):

    def generate_sequence(self) -> list[Command]:
        ...

    def build(self):
        ...


SequenceComponent = typing.Union[SequenceBuilder, Command]


class Serializer(Protocol):

    def serialize(self, data: typing.Union[dict, list]) -> str:
        ...


class Deserializer(Protocol):

    def deserialize(self, data: str) -> typing.Union[dict, list]:
        ...


class Logger(Protocol):

    def add_global_properties(self, properties: dict):
        ...

    def log_error(self, message: str, properties: dict = None):
        ...

    def log_exception(self, exception: Exception, properties: dict = None):
        ...

    def log_info(self, message: str, properties: dict = None):
        ...

    def log_event(self, message: str, properties: dict = None):
        ...

    def log_debug(self, message: str, properties: dict = None):
        ...

    def log_trace(self, message: str, properties: dict = None):
        ...



import base64
import inspect
import json
import re
import typing
from datetime import datetime
from decimal import Decimal
from enum import Enum

import dateutil

from formula_thoughts_web.abstractions import Serializer
from formula_thoughts_web.exceptions import MappingException


class LogSeverity(Enum):
    ERROR = "ERROR"
    EVENT = "EVENT"
    WARNING = "WARNING"
    INFO = "INFO"
    DEBUG = "DEBUG"
    TRACE = "TRACE"


class JsonConsoleLogger:

    def __init__(self, serializer: Serializer):
        self.__serializer = serializer
        self.__request_props = {}

    def __log(self, _type: LogSeverity, message: str, properties: dict):
        if properties is None:
            properties = {}
        _function = inspect.stack()[2].function
        _class = "None"
        _module = "Unknown"
        try:
            _class = type(inspect.stack()[2].frame.f_locals['self']).__name__
            _module = type(inspect.stack()[2].frame.f_locals['self']).__module__
        except:
            ...
        name = f"{_module}.{_class}.{_function}"
        message = {
            "message": message,
            "severity": str(_type),
            "location": name
        }
        log_message = {**message, **self.__request_props, **properties}
        print(f"{self.__serializer.serialize(log_message)}")

    def add_global_properties(self, properties: dict):
        self.__request_props = properties

    def log_error(self, message: str, properties: dict = None):
        self.__log(_type=LogSeverity.ERROR, message=message, properties=properties)

    def log_exception(self, exception: Exception, properties: dict = None):
        self.__log(_type=LogSeverity.ERROR, message=str(exception), properties=properties)

    def log_info(self, message: str, properties: dict = None):
        self.__log(_type=LogSeverity.INFO, message=message, properties=properties)

    def log_event(self, message: str, properties: dict = None):
        self.__log(_type=LogSeverity.EVENT, message=message, properties=properties)

    def log_debug(self, message: str, properties: dict = None):
        self.__log(_type=LogSeverity.DEBUG, message=message, properties=properties)

    def log_trace(self, message: str, properties: dict = None):
        self.__log(_type=LogSeverity.TRACE, message=message, properties=properties)


def all_annotations(cls):
    d = {}
    for c in cls.mro():
        try:
            d.update(**c.__annotations__)
        except AttributeError:
            # object, at least, has no __annotations__ attribute.
            pass
    return d


T = typing.TypeVar("T")


class ObjectMapper:

    def map_to_dict_and_ignore_none_fields(self, _from, to: typing.Type[T]) -> dict:
        print(f"{vars(_from).items()}")
        mapped = self.__generic_map(_from=_from,
                                    to=to,
                                    propValues=vars(_from).items())
        new_dict = mapped.__dict__
        self.__to_dict_and_ignore_none_fields(new_dict=new_dict, mapped=mapped)
        return new_dict

    def __to_dict_and_ignore_none_fields(self, new_dict: dict, mapped):
        for property, value in list(new_dict.items()):
            try:
                if bool(typing.get_type_hints(getattr(mapped, property))):
                    self.__to_dict_and_ignore_none_fields(
                        new_dict=value,
                        mapped=getattr(mapped, property))
            except TypeError:
                print(f"value {property} skipped")
            if value is None:
                new_dict.pop(property)

    def map(self, _from, to: typing.Type[T]) -> T:
        print(f"{vars(_from).items()}")
        return self.__generic_map(_from=_from,
                                  to=to,
                                  propValues=vars(_from).items(),
                                  map_to=lambda x, y: self.map(_from=x, to=y))

    def map_from_dict(self, _from, to: typing.Type[T]) -> T:
        print(_from.items())
        return self.__generic_map(_from=_from,
                                  to=to,
                                  propValues=_from.items(),
                                  map_to=lambda x, y: self.map_from_dict(_from=x, to=y))

    def map_to_dict(self, _from, to: typing.Type[T], preserve_decimal=False) -> dict:
        print(f"{vars(_from).items()}")
        return self.__generic_map(_from=_from,
                                  to=to,
                                  propValues=vars(_from).items(),
                                  map_callback=lambda x: self.to_dict(x, preserve_decimal=preserve_decimal),
                                  map_to=lambda x, y: self.map_to_dict(_from=x, to=y))

    def to_dict(self, obj, preserve_decimal):
        return json.loads(json.dumps(obj, default=lambda o: self.default_json_converter(o, preserve_decimal)))

    @staticmethod
    def default_json_converter(object, preserve_decimal):
        if type(object) == Decimal:
            if preserve_decimal:
                return str(object)
            else:
                return float(str(object))
        elif type(object) == datetime:
            return object.isoformat()
        else:
            return object.__dict__

    def __generic_map(self, _from, to, propValues, map_to, map_callback=lambda x: x):
        try:
            new_dto = to()
            dict_to = all_annotations(to)
            print("START MAPPING")
            print(f"all annotations from DTO {dict_to}")
            print(f"all props from _from {propValues}")
            for property, value in propValues:
                if property in dict_to:
                    if bool(typing.get_type_hints(dict_to[property])):
                        setattr(new_dto, property, map_callback(map_to(value, dict_to[property])))
                    elif (typing.get_origin(dict_to[property]) is list and
                          (bool(typing.get_type_hints(typing.get_args(dict_to[property])[0])))):
                        collection = []
                        sub_item_to = typing.get_args(dict_to[property])[0]
                        for item in value:
                            collection.append(map_callback(map_to(item, sub_item_to)))
                        setattr(new_dto, property, collection)
                    elif dict_to[property] is datetime and type(value) is str:
                        new_dto.__dict__[property] = datetime.fromisoformat(value)
                    elif dict_to[property] is Decimal:
                        if type(value) is str:
                            new_dto.__dict__[property] = Decimal(value)
                        elif type(value) is float:
                            new_dto.__dict__[property] = Decimal(str(value))
                        else:
                            new_dto.__dict__[property] = value
                    elif dict_to[property] == list[Decimal]:
                        decimals = []
                        for item in value:
                            if type(item) is str:
                                decimals.append(Decimal(item))
                            elif type(item) is float:
                                decimals.append(Decimal(str(item)))
                            else:
                                decimals.append(item)
                            new_dto.__dict__[property] = decimals
                    else:
                        new_dto.__dict__[property] = value
            print(f"__generic_map from {type(_from)} {to} and mapped {_from} out -> {new_dto}")
            return map_callback(new_dto)
        except Exception as e:
            raise MappingException(str(e))


class JsonSnakeToCamelSerializer:

    def serialize(self, data: typing.Union[dict, list]) -> str:
        return json.dumps(self.__snake_case_to_camel_case_dict(d=data), default=str)

    def __snake_case_to_camel_case_dict(self, d):
        if isinstance(d, list):
            return [self.__snake_case_to_camel_case_dict(i) if isinstance(i, (dict, list)) else self.__format_value(i)
                    for i in d]
        return {self.__snake_case_key_to_camel_case(a): self.__snake_case_to_camel_case_dict(b) if isinstance(b, (
            dict, list)) else self.__format_value(b) for a, b in d.items()}

    @staticmethod
    def __format_value(value) -> typing.Any:
        if (isinstance(value, Enum)):
            return value.value
        return value

    @staticmethod
    def __snake_case_key_to_camel_case(key: str) -> str:
        components = key.split('_')
        return components[0] + ''.join(x.title() for x in components[1:])


class JsonCamelToSnakeDeserializer:

    def deserialize(self, data: str) -> typing.Union[dict, list]:
        data_dict = json.loads(data)
        return self.__camel_case_to_snake_case_dict(d=data_dict)

    def __camel_case_to_snake_case_dict(self, d):
        if isinstance(d, list):
            return [self.__camel_case_to_snake_case_dict(i) if isinstance(i, (dict, list)) else i for i in d]
        return {self.__camel_case_key_to_snake_case(a): self.__camel_case_to_snake_case_dict(b) if isinstance(b, (
            dict, list)) else b for a, b in d.items()}

    @staticmethod
    def __camel_case_key_to_snake_case(key: str) -> str:
        words = re.findall(r'[A-Z]?[a-z]+|[A-Z]{1,}(?=[A-Z][a-z]|\d|\W|$)|\d+', key)
        return '_'.join(map(str.lower, words))


def utc_now() -> datetime:
    return datetime.utcnow()


def base64encode(s: str) -> str:
    return base64.b64decode(s).decode('utf-8')


def base64decode(s: str) -> str:
    return base64.b64encode(s.encode('utf-8')).decode('utf-8')


import inspect
from abc import ABC, abstractmethod

from formula_thoughts_web.abstractions import Logger, SequenceComponent, Command, SequenceBuilder, ApplicationContext, Error, \
    ErrorHandlingStrategy
from formula_thoughts_web.exceptions import StrategyNotFoundException

SUBSEQUENCE = "subsequence"
COMMAND = "command"


USE_RESPONSE_ERROR = "USE_RESPONSE_ERROR"
USE_EXCEPTION_ERROR = "USE_EXCEPTION_ERROR"


class FluentSequenceBuilder(ABC):

    def __init__(self):
        self.__components: list[(str, SequenceComponent)] = []

    def _add_command(self, command: Command) -> 'FluentSequenceBuilder':
        self.__components.append((COMMAND, command))
        return self

    def _add_sequence_builder(self, sequence_builder: SequenceBuilder) -> 'FluentSequenceBuilder':
        self.__components.append((SUBSEQUENCE, sequence_builder))
        return self

    def generate_sequence(self) -> list[Command]:
        self.build()
        new_list = []
        for (name, component) in self.__components:
            if name == COMMAND:
                new_list.append(component)
            if name == SUBSEQUENCE:
                sequence_component: SequenceBuilder = component
                commands = sequence_component.generate_sequence()
                new_list.extend(commands)
        return new_list

    @abstractmethod
    def build(self):
        ...

    @property
    def components(self) -> list[SequenceComponent]:
        return list(map(lambda x: x[1], self.__components))


class ErrorHandlingTypeState:

    def __init__(self, default_error_handling_strategy: str):
        self.__error_handling_type: str = default_error_handling_strategy

    @property
    def error_handling_type(self) -> str:
        return self.__error_handling_type

    @error_handling_type.setter
    def error_handling_type(self, value: str) -> None:
        self.__error_handling_type = value


class ExceptionErrorHandlingStrategy:

    def handle_error(self, context: ApplicationContext, error: Error) -> None:
        raise Exception(error.message)

    @property
    def strategy(self) -> str:
        return USE_EXCEPTION_ERROR


class ResponseErrorHandlingStrategy:

    def handle_error(self, context: ApplicationContext, error: Error) -> None:
        context.response = error

    @property
    def strategy(self) -> str:
        return USE_RESPONSE_ERROR


class ErrorHandlingStrategyFactory:

    def __init__(self, error_handling_strategies: list[ErrorHandlingStrategy],
                 error_handling_type_state: ErrorHandlingTypeState) -> None:
        self.__error_handling_type_state = error_handling_type_state
        self.__error_handling_strategies = error_handling_strategies

    def get_error_handling_strategy(self) -> ErrorHandlingStrategy:
        matching_strategies = list(filter(lambda x: x.strategy == self.__error_handling_type_state.error_handling_type, self.__error_handling_strategies))
        if len(matching_strategies) == 0:
            raise StrategyNotFoundException(f"{self.__error_handling_type_state.error_handling_type} not found")

        return matching_strategies[0]


class TopLevelSequenceRunner:

    def __init__(self, error_handling_strategy_factory: ErrorHandlingStrategyFactory, logger: Logger):
        self.__error_handling_strategy_factory = error_handling_strategy_factory
        self.__logger = logger

    def run(self, context: ApplicationContext,
            top_level_sequence: SequenceBuilder):
        commands = top_level_sequence.generate_sequence()
        for command in commands:
            name = "anonymous"
            try:
                name = f"{type(command).__module__}.{type(command).__name__}"
            except Exception:
                ...
            try:
                self.__logger.log_event(message="command event", properties={"action": name})
                self.__logger.log_info(f"begin command {name}")
                self.__logger.log_trace(f"request {context.body}")
                self.__logger.log_trace(f"response {context.response}")
                command.run(context)
                # for now, we throw on first error in top level sequence
                if any(context.error_capsules):
                    error = context.error_capsules[-1]
                    self.__logger.log_error(f"error found in error capsule {type(error).__name__}")
                    self.__error_handling_strategy_factory.get_error_handling_strategy().handle_error(context=context,
                                                                                                      error=error)
                    self.__logger.log_error(f"command pipeline SHORTED!")
                    break
            except Exception as e:
                self.__logger.log_error(f"command pipeline SHORTED due to an exception!")
                raise e
            finally:
                self.__logger.log_info(f"end command {name}")





from abc import ABC
from typing import Type

from formula_thoughts_web.abstractions import SequenceBuilder, ApplicationContext, ApiRequestHandler, Serializer, Logger, Deserializer
from formula_thoughts_web.application import TopLevelSequenceRunner, ErrorHandlingTypeState, USE_RESPONSE_ERROR
from formula_thoughts_web.crosscutting import ObjectMapper


class StatusCodeMapping:

    def __init__(self):
        self.__mappings = {}

    def add_mapping(self, _type, status_code: int) -> None:
        self.__mappings[f"{_type.__module__}.{_type.__name__}"] = status_code

    def get_mappings(self, response: Type) -> int:
        return self.__mappings[f"{response.__module__}.{response.__name__}"]


class WebRunner:

    def __init__(self,
                 request_handlers: list[ApiRequestHandler],
                 serializer: Serializer,
                 status_code_mappings: StatusCodeMapping,
                 error_handling_state: ErrorHandlingTypeState,
                 object_mapper: ObjectMapper,
                 logger: Logger):
        self.__object_mapper = object_mapper
        self.__error_handling_state = error_handling_state
        self.__status_code_mappings = status_code_mappings
        self.__logger = logger
        self.__serializer = serializer
        self.__request_handlers = request_handlers

    def run(self, event) -> dict:
        self.__error_handling_state.error_handling_type = USE_RESPONSE_ERROR
        headers = {"Content-Type": "application/json"}
        self.__logger.add_global_properties(properties={"route_key": event['routeKey']})
        request_handler_matches = list(filter(lambda x: x.route_key == event['routeKey'], self.__request_handlers))
        if len(request_handler_matches) == 0:
            return {
                "headers": headers,
                "body": self.__serializer.serialize(data={"message": f"route {event['routeKey']} not found"}),
                "statusCode": 404
            }
        try:
            context: ApplicationContext = request_handler_matches[0].run(event=event)
            body = None
            status_code = 204
            if context.response is not None:
                body = self.__serializer.serialize(data=self.__object_mapper.map_to_dict(_from=context.response, to=type(context.response)))
                status_code = self.__status_code_mappings.get_mappings(response=type(context.response))
            return {
                "headers": headers,
                "body": body,
                "statusCode": status_code
            }
        except Exception as e:
            self.__logger.log_exception(exception=e)
            return {
                "headers": headers,
                "body": self.__serializer.serialize(data={"message": f"internal server error :("}),
                "statusCode": 500
            }


class ApiRequestHandlerBase(ABC):

    def __init__(self, route_key: str,
                 sequence: SequenceBuilder,
                 command_pipeline: TopLevelSequenceRunner,
                 deserializer: Deserializer,
                 logger: Logger):
        self.__logger = logger
        self.__deserializer = deserializer
        self.__command_pipeline = command_pipeline
        self.__route_key = route_key
        self.__sequence = sequence

    def run(self, event: dict) -> ApplicationContext:
        body = None
        auth_user_id = None
        parameters = {}
        if 'pathParameters' in event:
            path_parameters = event['pathParameters']
            if path_parameters is None:
                pass
            parameters = {**parameters, **path_parameters}
            self.__logger.add_global_properties(properties={"path_params": path_parameters})
        if 'queryStringParameters' in event:
            query_parameters = event['queryStringParameters']
            if query_parameters is None:
                pass
            parameters = {**parameters, **query_parameters}
            self.__logger.add_global_properties(properties={"query_params": query_parameters})
        try:
            claims = auth_user_id = event['requestContext']['authorizer']['jwt']['claims']
            parameters = {**parameters, **claims}
            auth_user_id = claims['username']
            self.__logger.add_global_properties(properties={"auth_user_id": auth_user_id})
            self.__logger.add_global_properties(properties={"user_claims": claims})
        except KeyError:
            pass
        if 'body' in event:
            try:
                json_body = self.__deserializer.deserialize(event['body'])
                body = json_body
            except ValueError:
                self.__logger.log_info("Cannot serialize to dictionary, using string instead")
                body = event['body']
        context = ApplicationContext(body=body,
                                     auth_user_id=auth_user_id,
                                     variables=parameters,
                                     error_capsules=[])
        self.__command_pipeline.run(context=context,
                                    top_level_sequence=self.__sequence)
        return context

    @property
    def route_key(self) -> str:
        return self.__route_key



class EventSchemaInvalidException(Exception):
    pass


class EventNotFoundException(Exception):
    pass


class MappingException(Exception):
    pass


class StrategyNotFoundException(Exception):
    pass



from setuptools import setup, find_packages

DESCRIPTION = 'A web library for AWS lambda, gateway and SQS event handling'
LONG_DESCRIPTION = 'Allows for handling http api gateway requests, and SQS events'
REQUIREMENTS = [i.strip() for i in open("requirements.txt").readlines()]

# Setting up
setup(
    name="formula-thoughts-web",
    version="1.3.1",
    author="GanTheMan",
    author_email="aidanwilliamgannon@gmail.com",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=REQUIREMENTS,
    keywords=['python', 'lambda', 'api gateway', 'sqs'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix"
    ]
)


import typing
import uuid
from abc import ABC
from typing import Type

from botocore.client import BaseClient

from formula_thoughts_web.abstractions import SequenceBuilder, Deserializer, ApplicationContext, EventHandler, Logger, \
    Serializer
from formula_thoughts_web.application import TopLevelSequenceRunner, ErrorHandlingTypeState, USE_EXCEPTION_ERROR
from formula_thoughts_web.crosscutting import ObjectMapper
from formula_thoughts_web.exceptions import EventNotFoundException

EVENT = "EVENT"


class SQSEventPublisher:

    def __init__(self, sqs_client: BaseClient,
                 queue_name: str,
                 serializer: Serializer,
                 mapper: ObjectMapper):
        self.__mapper = mapper
        self.__serializer = serializer
        self.__sqs_client = sqs_client
        url = self.__sqs_client.get_queue_url(QueueName=queue_name)
        self.__queue_url = url["QueueUrl"]

    def send_sqs_message(self, message_group_id, payload: typing.Any):
        self.__sqs_client.send_message(
            QueueUrl=str(self.__queue_url),
            MessageBody=self.__serializer.serialize(data=self.__mapper.map_to_dict(_from=payload,
                                                                                   to=type(payload))),
            MessageGroupId=message_group_id,
            MessageAttributes={
                'messageType': {
                    'StringValue': type(payload).__name__,
                    'DataType': 'String'
                }
            },
            MessageDeduplicationId=str(uuid.uuid4())
        )


class EventRunner:

    def __init__(self, event_handlers: list[EventHandler],
                 error_handling_state: ErrorHandlingTypeState,
                 logger: Logger):
        self.__error_handling_state = error_handling_state
        self.__logger = logger
        self.__event_handlers = event_handlers

    def run(self, event: dict):
        failed_messages = []
        try:
            self.__error_handling_state.error_handling_type = USE_EXCEPTION_ERROR
            for message in event['Records']:
                try:
                    event_type = message['messageAttributes']['messageType']['stringValue']
                    body = message['body']
                    self.__logger.add_global_properties(properties={"event_type": event_type})
                    matching_handlers = list(filter(lambda x: f"{x.event_type.__name__}" == event_type, self.__event_handlers))
                    if len(matching_handlers) == 0:
                        raise EventNotFoundException(f"{event_type} does not match any found handlers")
                    matching_handlers[0].run(event=body)
                except Exception as e:
                    self.__logger.log_error(message="event runner captured exception")
                    self.__logger.log_exception(exception=e)
                    failed_messages.append(message["messageId"])
        except Exception as e:
            self.__logger.log_error(message="fatal error in event handler, retrying entire batch")
            self.__logger.log_exception(exception=e)
            raise
        return {
            'batchItemFailures': list(map(lambda x: {'itemIdentifier': x}, failed_messages))
        }


class EventHandlerBase(ABC):

    def __init__(self, event: Type,
                 sequence: SequenceBuilder,
                 command_pipeline: TopLevelSequenceRunner,
                 deserializer: Deserializer,
                 object_mapper: ObjectMapper):
        self.__object_mapper = object_mapper
        self.__deserializer = deserializer
        self.__command_pipeline = command_pipeline
        self.__sequence = sequence
        self.__event = event

    def run(self, event: str):
        event_dict = self.__deserializer.deserialize(event)
        event_object = self.__object_mapper.map_from_dict(_from=event_dict, to=self.__event)
        self.__command_pipeline.run(context=ApplicationContext(body=event_dict,
                                                               variables={EVENT: event_object},
                                                               error_capsules=[]),
                                    top_level_sequence=self.__sequence)

    @property
    def event_type(self) -> typing.Type:
        return self.__event
